@@ -1,3 +1,4 @@
+#!/usr/bin/perl
 # -*- perl -*-
 #
 #   DBD::File - A base class for implementing DBI drivers that
@@ -9,7 +10,7 @@
 #
 #  The original author is Jochen Wiedmann.
 #
-#  Copyright (C) 2009-2013 by H.Merijn Brand & Jens Rehsack
+#  Copyright (C) 2009-2020 by H.Merijn Brand & Jens Rehsack
 #  Copyright (C) 2004 by Jeff Zucker
 #  Copyright (C) 1998 by Jochen Wiedmann
 #
@@ -33,11 +34,10 @@ use warnings;
 
 use base qw( DBI::DBD::SqlEngine );
 use Carp;
-use vars qw( @ISA $VERSION $drh );
 
-$VERSION = "0.42";
+our $VERSION = "0.44";
 
-$drh = undef;		# holds driver handle(s) once initialized
+our $drh = undef;		# holds driver handle(s) once initialized
 
 sub driver ($;$)
 {
@@ -83,10 +83,10 @@ package DBD::File::dr;
 use strict;
 use warnings;
 
-use vars qw( @ISA $imp_data_size );
+use Carp;
 
-@DBD::File::dr::ISA           = qw( DBI::DBD::SqlEngine::dr );
-$DBD::File::dr::imp_data_size = 0;
+our @ISA           = qw( DBI::DBD::SqlEngine::dr );
+our $imp_data_size = 0;
 
 sub dsn_quote
 {
@@ -100,6 +100,35 @@ sub dsn_quote
 # XXX rewrite using TableConfig ...
 sub default_table_source { "DBD::File::TableSource::FileSystem" }
 
+sub connect
+{
+    my ($drh, $dbname, $user, $auth, $attr) = @_;
+
+    # We do not (yet) care about conflicting attributes here
+    # my $dbh = DBI->connect ("dbi:CSV:f_dir=test", undef, undef, { f_dir => "text" });
+    # will test here that both test and text should exist
+    #
+    # Parsing on our own similar to parse_dsn to find attributes in 'dbname' parameter.
+    if ($dbname) {
+	my $attr_hash = {
+	    map { (m/^\s* (\S+) \s*(?: =>? | , )\s* (\S*) \s*$/x) }
+	    split m/;/ => $dbname };
+	if (defined $attr_hash->{f_dir} && ! -d $attr_hash->{f_dir}) {
+	    my $msg = "No such directory '$attr_hash->{f_dir}";
+	    $drh->set_err (2, $msg);
+	    $attr_hash->{RaiseError} and croak $msg;
+	    return;
+	    }
+	}
+    if ($attr and defined $attr->{f_dir} && ! -d $attr->{f_dir}) {
+	my $msg = "No such directory '$attr->{f_dir}";
+	$drh->set_err (2, $msg);
+	return;
+	}
+
+    return $drh->SUPER::connect ($dbname, $user, $auth, $attr);
+    } # connect
+
 sub disconnect_all
 {
     } # disconnect_all
@@ -116,21 +145,19 @@ package DBD::File::db;
 use strict;
 use warnings;
 
-use vars qw( @ISA $imp_data_size );
-
 use Carp;
 require File::Spec;
 require Cwd;
 use Scalar::Util qw( refaddr ); # in CORE since 5.7.3
 
-@DBD::File::db::ISA           = qw( DBI::DBD::SqlEngine::db );
-$DBD::File::db::imp_data_size = 0;
+our @ISA           = qw( DBI::DBD::SqlEngine::db );
+our $imp_data_size = 0;
 
 sub data_sources
 {
     my ($dbh, $attr, @other) = @_;
     ref ($attr) eq "HASH" or $attr = {};
-    exists $attr->{f_dir}        or $attr->{f_dir}     = $dbh->{f_dir};
+    exists $attr->{f_dir}        or $attr->{f_dir}        = $dbh->{f_dir};
     exists $attr->{f_dir_search} or $attr->{f_dir_search} = $dbh->{f_dir_search};
     return $dbh->SUPER::data_sources ($attr, @other);
     } # data_source
@@ -270,10 +297,8 @@ package DBD::File::st;
 use strict;
 use warnings;
 
-use vars qw( @ISA $imp_data_size );
-
-@DBD::File::st::ISA           = qw( DBI::DBD::SqlEngine::st );
-$DBD::File::st::imp_data_size = 0;
+our @ISA           = qw( DBI::DBD::SqlEngine::st );
+our $imp_data_size = 0;
 
 my %supported_attrs = (
     TYPE      => 1,
@@ -294,12 +319,39 @@ sub FETCH
 
 	    # fill overall_defs unless we know
 	    unless (exists $sth->{f_overall_defs} && ref $sth->{f_overall_defs}) {
+		my $types = $sth->{Database}{Types};
+		unless ($types) { # Fetch types only once per database
+		    if (my $t = $sth->{Database}->type_info_all ()) {
+			foreach my $i (1 .. $#$t) {
+			    $types->{uc $t->[$i][0]}   = $t->[$i][1];
+			    $types->{$t->[$i][1]} ||= uc $t->[$i][0];
+			    }
+			}
+		    # sane defaults
+		    for ([  0, ""		],
+			 [  1, "CHAR"		],
+			 [  4, "INTEGER"	],
+			 [ 12, "VARCHAR"	],
+			 ) {
+			$types->{$_->[0]} ||= $_->[1];
+			$types->{$_->[1]} ||= $_->[0];
+			}
+		    $sth->{Database}{Types} = $types;
+		    }
 		my $all_meta =
 		    $sth->{Database}->func ("*", "table_defs", "get_sql_engine_meta");
-		while (my ($tbl, $meta) = each %$all_meta) {
+		foreach my $tbl (keys %$all_meta) {
+		    my $meta = $all_meta->{$tbl};
 		    exists $meta->{table_defs} && ref $meta->{table_defs} or next;
 		    foreach (keys %{$meta->{table_defs}{columns}}) {
-			$sth->{f_overall_defs}{$_} = $meta->{table_defs}{columns}{$_};
+			my $field_info = $meta->{table_defs}{columns}{$_};
+			if (defined $field_info->{data_type} &&
+				    $field_info->{data_type} !~ m/^[0-9]+$/) {
+			    $field_info->{type_name} = uc $field_info->{data_type};
+			    $field_info->{data_type} = $types->{$field_info->{type_name}} || 0;
+			    }
+			$field_info->{type_name} ||= $types->{$field_info->{data_type}} || "CHAR";
+			$sth->{f_overall_defs}{$_} = $field_info;
 			}
 		    }
 		}
@@ -307,7 +359,11 @@ sub FETCH
 	    my @colnames = $sth->sql_get_colnames ();
 
 	    $attr eq "TYPE"      and
-		return [ map { $sth->{f_overall_defs}{$_}{data_type}   || "CHAR" }
+		return [ map { $sth->{f_overall_defs}{$_}{data_type}   || 12 }
+			    @colnames ];
+
+	    $attr eq "TYPE_NAME" and
+		return [ map { $sth->{f_overall_defs}{$_}{type_name}   || "VARCHAR" }
 			    @colnames ];
 
 	    $attr eq "PRECISION" and
@@ -334,7 +390,7 @@ use warnings;
 
 use IO::Dir;
 
-@DBD::File::TableSource::FileSystem::ISA = "DBI::DBD::SqlEngine::TableSource";
+our @ISA = "DBI::DBD::SqlEngine::TableSource";
 
 sub data_sources
 {
@@ -343,6 +399,10 @@ sub data_sources
 	? $attr->{f_dir}
 	: File::Spec->curdir ();
     defined $dir or return; # Stream-based databases do not have f_dir
+    unless (-d $dir && -r $dir && -x $dir) {
+	$drh->set_err ($DBI::stderr, "Cannot use directory $dir from f_dir");
+	return;
+	}
     my %attrs;
     $attr and %attrs = %$attr;
     delete $attrs{f_dir};
@@ -421,7 +481,7 @@ use warnings;
 
 use Carp;
 
-@DBD::File::DataSource::Stream::ISA = "DBI::DBD::SqlEngine::DataSource";
+our @ISA = "DBI::DBD::SqlEngine::DataSource";
 
 # We may have a working flock () built-in but that doesn't mean that locking
 # will work on NFS (flock () may hang hard)
@@ -510,7 +570,7 @@ package DBD::File::DataSource::File;
 use strict;
 use warnings;
 
-@DBD::File::DataSource::File::ISA = "DBD::File::DataSource::Stream";
+our @ISA = "DBD::File::DataSource::Stream";
 
 use Carp;
 
@@ -531,7 +591,7 @@ sub complete_table_name
 	    }
 	}
 
-    # (my $tbl = $file) =~ s/$ext$//i;
+    # (my $tbl = $file) =~ s/\Q$ext\E$//i;
     my ($tbl, $basename, $dir, $fn_ext, $user_spec_file, $searchdir);
     if ($file_is_table and defined $meta->{f_file}) {
 	$tbl = $file;
@@ -540,10 +600,10 @@ sub complete_table_name
 	$user_spec_file = 1;
 	}
     else {
-	($basename, $dir, undef) = File::Basename::fileparse ($file, $ext);
+	($basename, $dir, undef) = File::Basename::fileparse ($file, qr{\Q$ext\E});
 	# $dir is returned with trailing (back)slash. We just need to check
-	# if it is ".", "./", or ".\"
-	if ($dir =~ m{^[.][/\\]?$} && ref $meta->{f_dir_search} eq "ARRAY") {
+	# if it is ".", "./", or ".\" or "[]" (VMS)
+	if ($dir =~ m{^(?:[.][/\\]?|\[\])$} && ref $meta->{f_dir_search} eq "ARRAY") {
 	    foreach my $d ($meta->{f_dir}, @{$meta->{f_dir_search}}) {
 		my $f = File::Spec->catdir ($d, $file);
 		-f $f or next;
@@ -611,12 +671,12 @@ sub complete_table_name
 	    }
 	@f > 0 && @f <= 2 and $file = $f[0];
 	!$respect_case && $meta->{sql_identifier_case} == 4 and # XXX SQL_IC_MIXED
-	    ($tbl = $file) =~ s/$ext$//i;
+	    ($tbl = $file) =~ s/\Q$ext\E$//i;
 
 	my $tmpfn = $file;
 	if ($ext && $req) {
             # File extension required
-            $tmpfn =~ s/$ext$//i or return;
+            $tmpfn =~ s/\Q$ext\E$//i or return;
             }
 	}
 
@@ -703,7 +763,7 @@ package DBD::File::Statement;
 use strict;
 use warnings;
 
-@DBD::File::Statement::ISA = qw( DBI::DBD::SqlEngine::Statement );
+our @ISA = qw( DBI::DBD::SqlEngine::Statement );
 
 # ====== SQL::TABLE ============================================================
 
@@ -719,7 +779,7 @@ require File::Spec;
 require Cwd;
 require Scalar::Util;
 
-@DBD::File::Table::ISA = qw( DBI::DBD::SqlEngine::Table );
+our @ISA = qw( DBI::DBD::SqlEngine::Table );
 
 # ====== UTILITIES ============================================================
 
@@ -894,6 +954,8 @@ sub DESTROY
     $meta->{lockfh} and $meta->{lockfh}->close ();
     undef $meta->{fh};
     undef $meta->{lockfh};
+
+    $self->SUPER::DESTROY();
     } # DESTROY
 
 1;
@@ -969,13 +1031,17 @@ C<< $sth->execute >>; undef for non-select statements.
 
 =head3 Unsupported DBI attributes and methods
 
-=head4 bind_param_inout
+=over 2
 
-=head4 CursorName
+=item bind_param_inout
 
-=head4 LongReadLen
+=item CursorName
 
-=head4 LongTruncOk
+=item LongReadLen
+
+=item LongTruncOk
+
+=back
 
 =head3 DBD::File specific attributes
 
@@ -994,6 +1060,11 @@ directory) when the dbh attribute is set.
 
   f_dir => "/data/foo/csv",
 
+If C<f_dir> is set to a non-existing location, the connection will fail.
+See CVE-2014-10401 for reasoning. Because of this, folders to use cannot
+be created after the connection, but must exist before the connection is
+initiated.
+
 See L<KNOWN BUGS AND LIMITATIONS>.
 
 =head4 f_dir_search
@@ -1161,12 +1232,12 @@ releases on
 
   @ary = DBI->data_sources ($driver);
   @ary = DBI->data_sources ($driver, \%attr);
-  
+
   @ary = $dbh->data_sources ();
   @ary = $dbh->data_sources (\%attr);
 
   @names = $dbh->tables ($catalog, $schema, $table, $type);
-  
+
   $sth = $dbh->table_info ($catalog, $schema, $table, $type);
   $sth = $dbh->table_info ($catalog, $schema, $table, $type, \%attr);
 
@@ -1366,7 +1437,7 @@ The original author is Jochen Wiedmann.
 
 =head1 COPYRIGHT AND LICENSE
 
- Copyright (C) 2009-2013 by H.Merijn Brand & Jens Rehsack
+ Copyright (C) 2009-2020 by H.Merijn Brand & Jens Rehsack
  Copyright (C) 2004-2009 by Jeff Zucker
  Copyright (C) 1998-2004 by Jochen Wiedmann
 
